"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SchemaParser = void 0;
const tslib_1 = require("tslib");
/**
 * 解析器是对输入的固定格式数据做拆解，使其符合引擎后续步骤预期，完成统一处理逻辑的步骤。
 * 本解析器面向的是标准 schema 协议。
 */
const change_case_1 = tslib_1.__importDefault(require("change-case"));
const types_1 = require("../types");
const const_1 = require("../const");
const errors_1 = require("../utils/errors");
const schema_1 = require("../utils/schema");
const common_1 = require("../utils/common");
const componentAnalyzer_1 = require("../analyzer/componentAnalyzer");
const validate_1 = require("../utils/validate");
const defaultContainer = {
    containerType: 'Component',
    componentName: 'Component',
    moduleName: 'Index',
    fileName: 'Index',
    css: '',
    props: {},
};
function getRootComponentName(typeName, maps) {
    if (maps[typeName]) {
        const rec = maps[typeName];
        if (rec.destructuring) {
            return rec.componentName || typeName;
        }
        const peerName = Object.keys(maps).find((depName) => {
            const depInfo = maps[depName];
            return (depName !== typeName &&
                !depInfo.destructuring &&
                depInfo.package === rec.package &&
                depInfo.version === rec.version &&
                depInfo.main === rec.main &&
                depInfo.exportName === rec.exportName &&
                depInfo.subName === rec.subName);
        });
        return peerName || typeName;
    }
    return typeName;
}
function processChildren(schema) {
    if (schema.props) {
        if (Array.isArray(schema.props)) {
            // FIXME: is array type props description
        }
        else {
            const nodeProps = schema.props;
            if (nodeProps.children) {
                if (!schema.children) {
                    // eslint-disable-next-line no-param-reassign
                    schema.children = nodeProps.children;
                }
                else {
                    let _children = [];
                    if (Array.isArray(schema.children)) {
                        _children = _children.concat(schema.children);
                    }
                    else {
                        _children.push(schema.children);
                    }
                    if (Array.isArray(nodeProps.children)) {
                        _children = _children.concat(nodeProps.children);
                    }
                    else {
                        _children.push(nodeProps.children);
                    }
                    // eslint-disable-next-line no-param-reassign
                    schema.children = _children;
                }
                delete nodeProps.children;
            }
        }
    }
}
function getInternalDep(internalDeps, depName) {
    const dep = internalDeps[depName];
    return (dep && dep.type !== types_1.InternalDependencyType.PAGE) ? dep : null;
}
class SchemaParser {
    validate(schema) {
        if (const_1.SUPPORT_SCHEMA_VERSION_LIST.indexOf(schema.version) < 0) {
            throw new types_1.CompatibilityError(`Not support schema with version [${schema.version}]`);
        }
        return true;
    }
    parse(schemaSrc) {
        // TODO: collect utils depends in JSExpression
        const compDeps = {};
        const internalDeps = {};
        let utilsDeps = [];
        const schema = this.decodeSchema(schemaSrc);
        // 解析三方组件依赖
        schema.componentsMap.forEach((info) => {
            var _a, _b;
            if (info.componentName) {
                compDeps[info.componentName] = Object.assign(Object.assign({}, info), { dependencyType: types_1.DependencyType.External, componentName: info.componentName, exportName: (_a = info.exportName) !== null && _a !== void 0 ? _a : info.componentName, version: info.version || '*', destructuring: (_b = info.destructuring) !== null && _b !== void 0 ? _b : false });
            }
        });
        let containers;
        // Test if this is a lowcode component without container  容器结构描述（IPublicTypeContainerSchema）
        if (schema.componentsTree.length > 0) {
            const firstRoot = schema.componentsTree[0];
            if (!firstRoot.fileName && !(0, schema_1.isValidContainerType)(firstRoot)) {
                // 整个 schema 描述一个容器，且无根节点定义
                const container = Object.assign(Object.assign(Object.assign({}, firstRoot), defaultContainer), { props: firstRoot.props || defaultContainer.props, css: firstRoot.css || defaultContainer.css, moduleName: firstRoot.moduleName || defaultContainer.moduleName, children: schema.componentsTree });
                containers = [container];
            }
            else {
                // 普通带 1 到多个容器的 schema
                containers = schema.componentsTree.map((n) => {
                    const subRoot = n;
                    const container = Object.assign(Object.assign({}, subRoot), { componentName: getRootComponentName(subRoot.componentName, compDeps), containerType: subRoot.componentName, moduleName: (0, validate_1.ensureValidClassName)(subRoot.componentName === "Component" /* ContainerType.Component */ ?
                            subRoot.fileName : change_case_1.default.pascalCase(subRoot.fileName)) });
                    return container;
                });
            }
        }
        else {
            throw new types_1.CodeGeneratorError("Can't find anything to generate.");
        }
        // 分析引用能力的依赖
        containers = containers.map((con) => (Object.assign(Object.assign({}, con), { analyzeResult: (0, componentAnalyzer_1.componentAnalyzer)(con) })));
        // 建立所有容器的内部依赖索引
        containers.forEach((container) => {
            let type;
            switch (container.containerType) {
                case 'Page':
                    type = types_1.InternalDependencyType.PAGE;
                    break;
                case 'Block':
                    type = types_1.InternalDependencyType.BLOCK;
                    break;
                default:
                    type = types_1.InternalDependencyType.COMPONENT;
                    break;
            }
            const dep = {
                type,
                moduleName: container.moduleName,
                destructuring: false,
                exportName: container.moduleName,
                dependencyType: types_1.DependencyType.Internal,
            };
            internalDeps[dep.moduleName] = dep;
        });
        const containersDeps = [].concat(...containers.map((c) => c.deps || []));
        // TODO: 不应该在出码部分解决？
        // 处理 children 写在了 props 里的情况
        containers.forEach((container) => {
            if (container.children) {
                // eslint-disable-next-line @typescript-eslint/no-invalid-void-type
                (0, schema_1.handleSubNodes)(container.children, {
                    node: (i) => processChildren(i),
                }, {
                    rerun: true,
                });
            }
        });
        containers.forEach((container) => {
            const depNames = this.getComponentNames(container);
            // eslint-disable-next-line no-param-reassign
            container.deps = (0, common_1.uniqueArray)(depNames, (i) => i)
                .map((depName) => getInternalDep(internalDeps, depName) || compDeps[depName])
                .filter(Boolean);
            // container.deps = Object.keys(compDeps).map((depName) => compDeps[depName]);
        });
        // 分析路由配置
        const routes = containers
            .filter((container) => container.containerType === 'Page')
            .map((page) => {
            const { meta } = page;
            if (meta) {
                return {
                    path: meta.router || `/${page.fileName}`,
                    fileName: page.fileName,
                    componentName: page.moduleName,
                };
            }
            return {
                path: '',
                fileName: page.fileName,
                componentName: page.moduleName,
            };
        });
        const routerDeps = routes
            .map((r) => internalDeps[r.componentName] || compDeps[r.componentName])
            .filter((dep) => !!dep);
        // 分析 Utils 依赖
        let utils;
        if (schema.utils) {
            utils = schema.utils;
            utilsDeps = schema.utils
                .filter((u) => u.type !== 'function')
                .map((u) => {
                var _a, _b;
                return (Object.assign(Object.assign({}, u.content), { componentName: u.name, version: u.content.version || '*', destructuring: (_a = u.content.destructuring) !== null && _a !== void 0 ? _a : false, exportName: (_b = u.content.exportName) !== null && _b !== void 0 ? _b : u.name }));
            });
        }
        else {
            utils = [];
        }
        // 分析项目 npm 依赖
        let npms = [];
        containers.forEach((con) => {
            const p = (con.deps || [])
                .map((dep) => {
                return dep.dependencyType === types_1.DependencyType.External ? dep : null;
            })
                .filter((dep) => dep !== null);
            const npmInfos = p.filter(Boolean).map((i) => ({
                package: i.package,
                version: i.version,
            }));
            npms.push(...npmInfos);
        });
        npms.push(...utilsDeps.map((utilsDep) => ({
            package: utilsDep.package,
            version: utilsDep.version,
        })));
        npms = (0, common_1.uniqueArray)(npms, (i) => i.package).filter(Boolean);
        return {
            containers,
            globalUtils: {
                utils,
                deps: utilsDeps,
            },
            globalI18n: schema.i18n,
            globalRouter: {
                routes,
                deps: routerDeps,
            },
            project: {
                css: schema.css,
                constants: schema.constants,
                config: schema.config || {},
                meta: schema.meta || {},
                i18n: schema.i18n,
                containersDeps,
                utilsDeps,
                packages: npms || [],
                dataSourcesTypes: this.collectDataSourcesTypes(schema),
                projectRemark: this.getProjectRemark(containers),
            },
        };
    }
    getProjectRemark(containers) {
        return {
            isSingleComponent: containers.length === 1 && containers[0].containerType === 'Component',
        };
    }
    getComponentNames(children) {
        return (0, schema_1.handleSubNodes)(children, {
            node: (i) => i.componentName,
        }, {
            rerun: true,
        });
    }
    decodeSchema(schemaSrc) {
        let schema;
        if (typeof schemaSrc === 'string') {
            try {
                schema = JSON.parse(schemaSrc);
            }
            catch (error) {
                throw new types_1.CodeGeneratorError(`Parse schema failed: ${(0, errors_1.getErrorMessage)(error) || 'unknown reason'}`);
            }
        }
        else {
            schema = schemaSrc;
        }
        return schema;
    }
    collectDataSourcesTypes(schema) {
        var _a, _b;
        const dataSourcesTypes = new Set();
        // 数据源的默认类型为 fetch
        const defaultDataSourceType = 'fetch';
        // 收集应用级别的数据源
        (_b = (_a = schema.dataSource) === null || _a === void 0 ? void 0 : _a.list) === null || _b === void 0 ? void 0 : _b.forEach((ds) => {
            dataSourcesTypes.add(ds.type || defaultDataSourceType);
        });
        // 收集容器级别的数据源（页面/组件/区块）
        schema.componentsTree.forEach((rootNode) => {
            var _a, _b;
            (_b = (_a = rootNode.dataSource) === null || _a === void 0 ? void 0 : _a.list) === null || _b === void 0 ? void 0 : _b.forEach((ds) => {
                dataSourcesTypes.add(ds.type || defaultDataSourceType);
            });
        });
        return Array.from(dataSourcesTypes.values());
    }
}
exports.SchemaParser = SchemaParser;
exports.default = SchemaParser;
//# sourceMappingURL=SchemaParser.js.map